/*----------------------------------------------------------------------------

	atreevw.cpp
	Copyright 1997 Sumus Computing Incorporated, All Rights Reserved.
	
	Beta 0.1, 97/02/02

----------------------------------------------------------------------------*/

#include "atreevw.h"


//----------------------------------------------------------------------------

BBitmap* g_NewBitmap(long width,long height,void const* data)
{
	BRect bitmaprect(0,0,width-1,height-1);

	BBitmap *pb = new BBitmap(bitmaprect,B_COLOR_8_BIT);
	pb->SetBits(data,pb->BitsLength(),0,B_COLOR_8_BIT);
	
	return pb;
}


BBitmap* g_NewBitmap(unsigned char const *data)
{
	return g_NewBitmap(data[0],data[1],data+2);
}


enum e8BitColors
{
	cWW = 0xFF,	// white
	cBB = 0x00,	// black
	cRR = 0x2A,	// red

	cDB = 0x25,	// dark blue
	cLB = 0x66,	// light blue

	cDG = 0x36,	// dark gray
	cLG = 0x46,	// light gray

	cYY = 0xE5	// yellow
};


static unsigned char const dataBITMAP_COLLAPSED[] = 
{
	6,7,
	cWW,cBB,cBB,cWW,cWW,cWW,0,0,
	cWW,cBB,cLB,cBB,cWW,cWW,0,0,
	cWW,cBB,cLB,cLB,cBB,cWW,0,0,
	cWW,cBB,cLB,cLB,cLB,cBB,0,0,
	cWW,cBB,cLB,cLB,cBB,cWW,0,0,
	cWW,cBB,cLB,cBB,cWW,cWW,0,0,
	cWW,cBB,cBB,cWW,cWW,cWW,0,0
};


static unsigned char const dataBITMAP_EXPANDED[] = 
{
	7,5,
	cBB,cBB,cBB,cBB,cBB,cBB,cBB,0,
	cBB,cLB,cLB,cLB,cLB,cLB,cBB,0,
	cWW,cBB,cLB,cLB,cLB,cBB,cWW,0,
	cWW,cWW,cBB,cLB,cBB,cWW,cWW,0,
	cWW,cWW,cWW,cBB,cWW,cWW,cWW,0
};


static unsigned char const dataBITMAP_DEFAULT_ICON[] = 
{
	5,5,
	cWW,cBB,cBB,cBB,cWW,0,0,0,
	cBB,cRR,cRR,cRR,cBB,0,0,0,
	cBB,cRR,cRR,cRR,cBB,0,0,0,
	cBB,cRR,cRR,cRR,cBB,0,0,0,
	cWW,cBB,cBB,cBB,cWW,0,0,0
};


//----------------------------------------------------------------------------

static void DrawCenterBitmap(BView& view,BBitmap const* image,BPoint point)
{
	BRect br = image->Bounds();
		
	point.x -= br.Width() / 2.0;
	point.y -= br.Height() / 2.0;

	view.DrawBitmap(image,point);
}


//----------------------------------------------------------------------------

class AWindowLock
{
public:
	AWindowLock(BWindow* a_window)
	  :	window(a_window)
	{
		if(window) window->Lock();
	}

	~AWindowLock()
	{
		if(window) window->Unlock();
	}

private:
	BWindow *window;
};


#define WITH_LOCK(code)						\
	{										\
		AWindowLock windowLock(Window());	\
		code								\
	}

	
//----------------------------------------------------------------------------

ATreeItem::ATreeItem(long itemsPerBlock)
  :	BList(itemsPerBlock),
 	parent(NULL), isCollapsed(false)
{
}


ATreeItem::~ATreeItem()
{
}


bool ATreeItem::AddItem(ATreeItem& item,long atIndex)
{
	WITH_LOCK
	(
		if(NULL!=item.parent) 
			return false;					// already inserted in a tree

		if(!inherited::AddItem(&item,atIndex))
			return false;					// BList AddItem failed
	
		item.parent = this;
		Notify(notifyADDITEM,item,VisibleIndex(&item));
		Notify(notifyREFRESH,*this,0);
	)
	
	return true;
}


bool ATreeItem::RemoveItem(ATreeItem& item)
{
	WITH_LOCK
	(
		if(this!=item.parent) 
			return false;					// not this item's child
	
		Notify(notifyREMOVEITEM,item,VisibleIndex(&item));
		inherited::RemoveItem(&item);		// always true	//>>> VERIFY
		Notify(notifyREFRESH,*this,0);
		item.parent = NULL;
	)
	
	return true;
}


ATreeItem* ATreeItem::RemoveItem(long atIndex)
{
	WITH_LOCK
	(
		ATreeItem *item = ItemAt(atIndex);

		if(NULL==item || !RemoveItem(*item))
			return NULL;

		return item;
	)
}


void ATreeItem::MakeEmpty(bool deleteItems)
{
	WITH_LOCK
	(
		while(0 < CountItems())
		{
			ATreeItem *item = RemoveItem(0);
			if(deleteItems) delete item;
		}
	)
}


void ATreeItem::MakeEmptyAll(bool deleteItems)
{
	WITH_LOCK
	(
		while(0 < CountItems())
		{
			ATreeItem *item = RemoveItem(0);	// never NULL
			item->MakeEmptyAll(deleteItems);
			if(deleteItems) delete item;
		}
	)
}


void ATreeItem::Collapse()
{
	WITH_LOCK
	(
		if(isCollapsed) return;

		for(long l=CountItems()-1; l>=0; l--)
		{
			ATreeItem *item = ItemAtFast(l);	// never NULL
			Notify(notifyREMOVEITEM,*item,VisibleIndex(item));
		}

		isCollapsed = true;
		Notify(notifyREFRESH,*this,0);
	)
}


void ATreeItem::Expand()
{
	WITH_LOCK
	(
		if(!isCollapsed) return;
		isCollapsed = false;

		for(long l=0,lCount=CountItems(); l<lCount; l++)
		{
			ATreeItem *item = ItemAtFast(l);	// never NULL
			Notify(notifyADDITEM,*item,VisibleIndex(item));
		}

		Notify(notifyREFRESH,*this,0);
	)
}


void ATreeItem::CollapseAll()
{
	WITH_LOCK
	(
		Collapse();

		for(long l=0,lCount=CountItems(); l<lCount; l++)
		{
			ItemAtFast(l)->CollapseAll();
		}
	)
}


void ATreeItem::ExpandAll()
{
	WITH_LOCK
	(
		Expand();

		for(long l=0,lCount=CountItems(); l<lCount; l++)
		{
			ItemAtFast(l)->ExpandAll();
		}
	)
}


bool ATreeItem::CanBeCollapsed() const
{
	return !isCollapsed && !IsEmpty();
}


bool ATreeItem::CanBeExpanded() const
{
	return isCollapsed && !IsEmpty();
}


long ATreeItem::Layer() const
{
	return parent ? 1 + parent->Layer() : 0;		
}


long ATreeItem::VisibleIndex(ATreeItem* item) const
{
	long lCount;
	
	if(isCollapsed)
	{
		lCount = 0;
	}
	else
	{
		if(NULL==item)
		{
			lCount = CountItems();
		}
		else
		{
			lCount = IndexOf(*item);	
			if(B_ERROR==lCount)
				return B_ERROR;
		}
	}

	long ix = 1;							// 1 for this item
	for(long l=0; l<lCount; l++)
		ix += ItemAt(l)->VisibleIndex();	// plus children

	return ix;
}


void ATreeItem::Draw(BView& view,BRect rect)
{
	long layer = Layer();
	
	DrawLink(view,rect,layer);
	DrawIcon(view,rect,layer);
	DrawLabel(view,rect,layer);
}


BBitmap const* ATreeItem::GetDefaultIcon() const
{
	return parent ? parent->GetDefaultIcon() : NULL;
}


static float const DX = 12.0;					// horizontal distance

void ATreeItem::DrawLink(BView& view,BRect rect,long layer)
{
	float const HY = (rect.top+rect.bottom)/2;	// vertical center
	float X = DX*(layer-0.9), XX=X;				// horizontal position

	// draw a little angle
	if(layer>1)
	{
		view.StrokeLine(BPoint(X,HY),BPoint(X+DX/2,HY), B_MIXED_COLORS);
		view.StrokeLine(BPoint(X,HY),BPoint(X,rect.top),B_MIXED_COLORS);
	}
	
	// draw connecting lines
	ATreeItem *item=this, *p=parent;
	while(layer>1 && NULL!=p)
	{
		if(item!=p->LastItem())
		{
			view.StrokeLine(BPoint(XX,rect.top),BPoint(XX,rect.bottom),B_MIXED_COLORS);
		}
		
		XX -= DX;	layer--;
		item = p;	p = p->parent;
	}
	
	// draw the blue triangle
	BBitmap const *linkIcon = NULL;

	if(CanBeExpanded())
	{
		linkIcon = ATreeView::bitmapCollapsed;
	}
	else if(CanBeCollapsed())
	{
		linkIcon = ATreeView::bitmapExpanded;
	}

	if(NULL!=linkIcon)
		DrawCenterBitmap(view,linkIcon,BPoint(X+DX/3,HY));
}


void ATreeItem::DrawIcon(BView& view,BRect rect,long layer)
{
	BBitmap const* icon = GetIcon();

	if(icon)
	{
		float const HX = DX*(layer+0.1);
		float const HY = (rect.top+rect.bottom)/2;
		DrawCenterBitmap(view,icon,BPoint(HX,HY));
	}
}


void ATreeItem::DrawLabel(BView& view,BRect rect,long layer)
{
	char const* label = GetLabel();

	if(label)
	{
		float const X  = DX*(layer+1);
		view.DrawString(label,BPoint(X,rect.bottom));
	}
}


bool ATreeItem::Notify(eNotify msg,ATreeItem& item,long index)
{
	if(NULL==parent) return false;

	switch(msg)
	{
	case notifyADDITEM:
	case notifyREMOVEITEM:
		if(!isCollapsed)
		{
			index += parent->VisibleIndex(this);
			return parent->Notify(msg,item,index);
		}
		break;

	case notifyREFRESH:
		if(!parent->isCollapsed)
		{
			index += parent->VisibleIndex(this);
			return parent->Notify(msg,item,index);
		}
		break;
	}
	
	return false;
}


BWindow* ATreeItem::Window() const
{
	return NULL!=parent ? parent->Window() : NULL;
}


//----------------------------------------------------------------------------

ATreeView::ARootItem::ARootItem()
  :	treeView(NULL)
{
}


BBitmap const* ATreeView::ARootItem::GetDefaultIcon() const
{
	return ATreeView::bitmapDefaultIcon;
}


bool ATreeView::ARootItem::Notify(ATreeItem::eNotify msg,ATreeItem& item,long index)
{
	long atIndex = index-1;			// discount :)
	if(atIndex<0) return false;

	switch(msg)
	{
	case ATreeItem::notifyADDITEM:
		treeView->NotifyAddItem(item,atIndex);
		return true;
		
	case ATreeItem::notifyREMOVEITEM:
		treeView->NotifyRemoveItem(atIndex);
		return true;

	case ATreeItem::notifyREFRESH:
		treeView->NotifyRefresh(atIndex);
		return true;
	}
	
	return false;
}


BWindow* ATreeView::ARootItem::Window() const
{
	return treeView->Window();
}


static ulong const MSG_EXPAND		= 'tvxx';
static ulong const MSG_COLLAPSE		= 'tvcc';
static ulong const MSG_EXPAND_ALL	= 'tvxa';
static ulong const MSG_COLLAPSE_ALL	= 'tvca';


ATreeView::ATreeView(BRect frame,char const* name,
					 ulong resizingMode,ulong flags)
  :	BListView(frame,name,resizingMode,flags)
{
	rootItem.treeView = this;
	
	if(0==instanceCount++)
	{
		bitmapCollapsed		= g_NewBitmap(dataBITMAP_COLLAPSED);
		bitmapExpanded		= g_NewBitmap(dataBITMAP_EXPANDED);
		bitmapDefaultIcon	= g_NewBitmap(dataBITMAP_DEFAULT_ICON);

		popupMenu = new BPopUpMenu("");
		popupMenu->AddItem(new BMenuItem("Expand",new BMessage(MSG_EXPAND)));
		popupMenu->AddItem(new BMenuItem("Collapse",new BMessage(MSG_COLLAPSE)));
		popupMenu->AddItem(new BMenuItem("Expand All",new BMessage(MSG_EXPAND_ALL)));
		popupMenu->AddItem(new BMenuItem("Collapse All",new BMessage(MSG_COLLAPSE_ALL)));
	}
}


ATreeView::~ATreeView()
{
	if(0==--instanceCount)
	{
		delete popupMenu;			popupMenu			= NULL;
		delete bitmapCollapsed;		bitmapCollapsed		= NULL;
		delete bitmapExpanded;		bitmapExpanded		= NULL;
		delete bitmapDefaultIcon;	bitmapDefaultIcon	= NULL;
	}
}


ulong ATreeView::instanceCount				= 0;

BPopUpMenu *ATreeView::popupMenu			= NULL;

BBitmap const* ATreeView::bitmapCollapsed	= NULL;
BBitmap const* ATreeView::bitmapExpanded	= NULL;
BBitmap const* ATreeView::bitmapDefaultIcon	= NULL;


bool ATreeView::Lock()
{
	BWindow *window = Window();
	return window ? window->Lock() : false;
}


void ATreeView::Unlock()
{
	BWindow *window = Window();
	if(window) window->Unlock();
}


void ATreeView::DrawItem(BRect updateRect,long index)
{
	BRect rect = ItemFrame(index);
	
	if(updateRect.Intersects(rect))
	{
		ATreeItem* item = static_cast<ATreeItem*>(ItemAt(index)); 	// not NULL
		WITH_LOCK( item->Draw(*this,rect);)
	}
}


void ATreeView::MouseDown(BPoint point)
{
	inherited::MouseDown(point);

	long index = CurrentSelection();
	if(index<0) return;
		
	ATreeItem *item = static_cast<ATreeItem*>(ItemAt(index));		// not NULL

	long buttons = Window()->CurrentMessage()->FindLong( "buttons");
	if(buttons & 2)	// right click
	{
		EnableMenuItem(MSG_EXPAND,		item->CanBeExpanded());
		EnableMenuItem(MSG_EXPAND_ALL,	item->CanBeExpanded());
		EnableMenuItem(MSG_COLLAPSE,	item->CanBeCollapsed());
		EnableMenuItem(MSG_COLLAPSE_ALL,item->CanBeCollapsed());

		ConvertToScreen(&point);
		BMenuItem *selected = popupMenu->Go(point);

		if(selected)
		{
			switch(selected->Message()->what)
			{
			case MSG_EXPAND:
				item->Expand();
				break;
			case MSG_COLLAPSE:
				item->Collapse();
				break;
			case MSG_EXPAND_ALL:
				item->ExpandAll();
				break;
			case MSG_COLLAPSE_ALL:
				item->CollapseAll();
				break;
			}
		}
	}
	else			// not a right click
	{
		BRect rect = ItemFrame(index);
	
		float       X = DX*(item->Layer() - 0.9) + DX/3;
		float const HY = (rect.top+rect.bottom)/2;
		float const D = 4;
		
		// blue triangle clicked	//>>>todo
		if(BRect(X-D,HY-D,X+D,HY+D).Contains(point))
		{			
			if(item->IsCollapsed())
				item->Expand();
			else
				item->Collapse();
		}
	}
}


void ATreeView::KeyDown(ulong aChar)
{
	switch(aChar)
	{
	case '+':
	case '-':
		{
			long index = CurrentSelection();
			if(index<0) break;
		
			ATreeItem *item = static_cast<ATreeItem*>(ItemAt(index));	// not NULL

			if('+'==aChar)
				item->Expand();
			else
				item->Collapse();
		}
		break;

	default:
		inherited::KeyDown(aChar);
		break;
	}
}


void ATreeView::NotifyAddItem(ATreeItem& item,long& atIndex)
{
	AddItem(&item,atIndex);

	// redraw preceding visible items, to draw the connecting line
	long layer=item.Layer();
	for(long l=atIndex-1; l>=0; l--)
	{
		ATreeItem *item1 = static_cast<ATreeItem*>(ItemAt(l));	// not NULL
		NotifyRefresh(l);
		if(layer>=item1->Layer()) 
			break;
	} 
	
	// if the item is expanded, recursively add its children
	atIndex++;
	if(!item.isCollapsed)
	{
		for(long l=0,lCount=item.CountItems(); l<lCount; l++)
		{
			NotifyAddItem(*item.ItemAtFast(l),atIndex);
		}
	}
}


void ATreeView::NotifyRemoveItem(long atIndex)
{
	ATreeItem *item = static_cast<ATreeItem*>(ItemAt(atIndex));	// not NULL
	
	for(long l=0,lCount=item->VisibleIndex(); l<lCount; l++)
	{
		RemoveItem(atIndex);
	}
}


void ATreeView::NotifyRefresh(long atIndex)
{
	InvalidateItem(atIndex);
}


void ATreeView::EnableMenuItem(ulong msg,bool enable)
{
	BMenuItem *item = popupMenu->FindItem(msg);
	if(item) item->SetEnabled(enable);
}


//-- eof ---------------------------------------------------------------------
