/*----------------------------------------------------------------------------

	atreevw.h
	Copyright 1997 Sumus Computing Incorporated, All Rights Reserved.

	Beta 0.1, 97/02/02
	
----------------------------------------------------------------------------*/

//@Man: TreeView
/**	A treeview is a view that displays a hierarchical list of items.
*/

#ifndef __ATREEVW_H__
#define __ATREEVW_H__

//----------------------------------------------------------------------------

/// creates a new bitmap (width/height; data: one byte per pixel, pad to 8*x B)
BBitmap* g_NewBitmap(long width,long height,void const* data);

/*@ManMemo:
	creates a new bitmap; data: 1st byte is a bitmap width, 
	2nd byte is a bitmap height, rest is one byte per pixel, pad to 8*x B
*/
BBitmap* g_NewBitmap(unsigned char const *data);

//----------------------------------------------------------------------------

/// a tree item base class
/**	ATreeItem is a node (root, node or leaf) in a tree sructure. Derived from 
	BList, contains pointers to child items. ATreeItem is never instantiated 
	directly.
	An application should derive a new class from ATreeItem, make the
	constructor public, and typically add data members and override GetLabel()
	and GetIcon(). Some of BList's member functions are typed and exported as 
	public members in ATreeItem.
	Note that pointers used in BList method are sometimes changed to references 
	in ATreeItem methods -- the idea is that NULL could be passed as a pointer
	value; on the other hand, a reference should always refer to an object. 
*/
class ATreeItem: protected BList
{
	friend class ATreeView;

protected:
	//@Man:	constructor
	//@{
	/// this class is not meant to be directly instantiated
	ATreeItem(long itemsPerBlock = 20);
	//@}

private:
	//@Man:	disabled methods
	//@{
	/// copy constructor is not available in this version
	ATreeItem(ATreeItem const&);
	/// assignment is not available in this version
	ATreeItem& operator=(ATreeItem const& item);
	//@}

public:
	//@Man:	destructor
	//@{
	///
	virtual ~ATreeItem();
	//@}

public:
	//@Man:	typed BList-like methods
	//@{
	///
	long CountItems() const;
	///
	bool AddItem(ATreeItem& item);
	///
	bool AddItem(ATreeItem& item,long atIndex);
	///
	bool RemoveItem(ATreeItem& item);
	///
	ATreeItem* RemoveItem(long index);
	///
	ATreeItem* ItemAt(long index) const;
	///
	ATreeItem* ItemAtFast(long index) const;
	///
	long IndexOf(ATreeItem& item) const;
	///
	ATreeItem* FirstItem() const;
	///
	ATreeItem* LastItem() const;
	///
	bool HasItem(ATreeItem& item) const;
	///
	bool IsEmpty() const;
	///
	void DoForEach(bool (*func)(ATreeItem&));
	///
	void DoForEach(bool (*func)(ATreeItem&,void*),void*);
	///
	void SortItems(int (*cmp)(ATreeItem const&,ATreeItem const&));
	/// removes all child items; optionally deletes them
	void MakeEmpty(bool deleteItems = FALSE);
	/// recursive MakeEmpty
	void MakeEmptyAll(bool deleteItems = FALSE);
	//@}
	
private:
	//@Man:	typed BList methods, not yet implemented
	//@{
	///
	bool AddList(ATreeItem& newItems);
	///
	bool AddList(ATreeItem& newItems,long atIndex);
	//@}

public:
	//@Man: collapsing, expanding
	//@{
	/// collapses the item; overriden version may remove the child items
	virtual void Collapse();
	/// recursive Collapse()
	void CollapseAll();
	/// expands the item; overriden version may insert child items
	virtual void Expand();
	/// recursive Expand()
	void ExpandAll();
	/// returns the collapse status of the item
	bool IsCollapsed() const;
	/// returns !IsEmpty() && !IsCollapsed()
	virtual bool CanBeCollapsed() const;
	/// returns !IsEmpty() && IsCollapsed()
	virtual bool CanBeExpanded() const;
	//@}

	//@Man:	methods typically overriden in derived classes
	//@{
	/// string label, displayed in the treeview
	virtual char const* GetLabel() const;
	/// icon displayed in the treeview
	virtual BBitmap const* GetIcon() const;
	//@}
	
protected:
	//@Man:	implementation details
	//@{
	/// returns the layer of the item; 0=the very root
	long Layer() const;
	/*ManMemo:
		Returns an index of the child item; only visible (not collapsed) 
		items count. If item==NULL, returns total number of visible items.
	*/
	long VisibleIndex(ATreeItem* item = NULL) const;
	/// draws the item; calls DrawLink(), DrawIcon(), DrawLabel()
	void Draw(BView& view,BRect rect);
	//@}

	//@Man:	overridable implementation details
	//@{
	/// returns the default icon
	virtual BBitmap const* GetDefaultIcon() const;
	/// 
	virtual void DrawLink(BView& view,BRect rect,long layer);
	///
	virtual void DrawIcon(BView& view,BRect rect,long layer);
	///
	virtual void DrawLabel(BView& view,BRect rect,long layer);
	//@}

	//@Man:	implementation details, not to be overriden
	//@{
	enum eNotify
	{
		notifyADDITEM,notifyREMOVEITEM,notifyREFRESH
	};
	///
	virtual bool Notify(eNotify msg,ATreeItem& item,long index);
	///
	virtual BWindow* Window() const;
	//@}
	
private:
	ATreeItem	*parent;
	bool		isCollapsed;
};


inline long ATreeItem::CountItems() const
{
	return inherited::CountItems();
}


inline bool ATreeItem::AddItem(ATreeItem& item)
{
	return AddItem(item,CountItems());
}


inline ATreeItem* ATreeItem::ItemAt(long index) const
{
	return static_cast<ATreeItem*>(inherited::ItemAt(index));
}


inline ATreeItem* ATreeItem::ItemAtFast(long index) const
{
	return static_cast<ATreeItem*>(inherited::ItemAtFast(index));
}


inline long ATreeItem::IndexOf(ATreeItem& item) const
{
	return inherited::IndexOf(&item);
}


inline ATreeItem* ATreeItem::FirstItem() const
{
	return static_cast<ATreeItem*>(inherited::FirstItem());
}


inline ATreeItem* ATreeItem::LastItem() const
{
	return static_cast<ATreeItem*>(inherited::LastItem());
}


inline bool ATreeItem::HasItem(ATreeItem& item) const
{
	return inherited::HasItem(&item);
}


inline bool ATreeItem::IsEmpty() const
{
	return inherited::IsEmpty();
}


inline void ATreeItem::DoForEach(bool (*func)(ATreeItem&))
{
	inherited::DoForEach(reinterpret_cast<bool(*)(void*)>(func));
}


inline void ATreeItem::DoForEach(bool (*func)(ATreeItem&,void*),void* arg2)
{
	inherited::DoForEach(reinterpret_cast<bool(*)(void*,void *)>(func),arg2);
}


inline void ATreeItem::SortItems(int (*cmp)(ATreeItem const&,ATreeItem const&))
{
	inherited::SortItems(reinterpret_cast<int(*)(void const*,void const*)>(cmp));
}


inline bool ATreeItem::IsCollapsed() const
{
	return isCollapsed;
}


inline char const* ATreeItem::GetLabel() const
{
	return NULL;
}


inline BBitmap const* ATreeItem::GetIcon() const
{
	return GetDefaultIcon();
}


//----------------------------------------------------------------------------

/// a tree view class
/**	ATreeView is a class derived from BListView. It contains one instance of 
	the ARootItem (see treeRoot), which is the very root of a tree to be 
	displayed. The treeRoot itself is not visible. Every its child item is 
	a visible root of a subtree. This way the view can contain several visible 
	trees, while maintaining only one.
*/
class ATreeView: protected BListView
{
public:
	/// a root item class
	class ARootItem: public ATreeItem
	{
		friend class ATreeView;

	public:
		///
		ARootItem();
		///
		virtual BBitmap const* GetDefaultIcon() const;

	protected:
		//@Man:	implementation details
		//@{
		///
		virtual bool Notify(ATreeItem::eNotify msg,ATreeItem& item,long index);
		///
		virtual BWindow* Window() const;
		//@}

	private:
		ATreeView* treeView;
	};
	
	friend class ATreeItem;
	friend class ATreeView::ARootItem;

public:
	/// constructor, for description see BListView
	ATreeView(BRect frame,
			  char const* name,
			  ulong resizingMode = B_FOLLOW_LEFT|B_FOLLOW_TOP, 
			  ulong flags		 = B_WILL_DRAW|B_NAVIGABLE|B_FRAME_EVENTS);
	///
	virtual ~ATreeView();
	
	/// returns a reference to the very root item
	ARootItem& Root();
	/// locks the (attached to) window
	bool Lock();
	/// unlocks the (attached to) window
	void Unlock(); 
	
protected:
	/// for description see BListView
	virtual void DrawItem(BRect updateRect,long index);
	/// as in BListView, right click brings up a popup menu
	virtual void MouseDown(BPoint point);
	/// +, -
	virtual void KeyDown(ulong aChar);

	//@Man:	implementation details
	//@{
	///
	void NotifyAddItem(ATreeItem& item,long& atIndex);
	///
	void NotifyRemoveItem(long atIndex);
	///
	void NotifyRefresh(long atIndex);
	//@}

private:
	static ulong instanceCount;

	ARootItem rootItem;

	static BPopUpMenu 		*popupMenu;
	static void EnableMenuItem(ulong msg,bool enable);

	static BBitmap const	*bitmapCollapsed,
							*bitmapExpanded,
							*bitmapDefaultIcon;
};


inline ATreeView::ARootItem& ATreeView::Root()
{
	return rootItem;
}


//----------------------------------------------------------------------------

#endif


//-- eof ---------------------------------------------------------------------
